@@ -43,7 +43,7 @@ from pint import UnitRegistry, Quantity
 from pint.unit import UnitDefinition
 from pint.converters import ScaleConverter
 from pint.errors import DimensionalityError
-
+from datetime import datetime
 
 MAX_WORKERS = 10
 MAX_ELEMS = 10
@@ -62,6 +62,7 @@ SUPPORTED_MIMES = [t().mime for t in SUPPORTED_FILETYPES]
 
 j2h = Json2Html()
 pd.options.plotting.backend = "plotly"
+pd.set_option('mode.use_inf_as_na', True)
 pio.templates.default = "simple_white"
 warnings.formatwarning = lambda msg, *args, **kwargs: f"{msg}\n"
 warnings.filterwarnings("default", category=DeprecationWarning, module=__name__)
@@ -153,15 +154,23 @@ def visit(path, key, value):
 
 
 class Dict(dict):
-    def pretty(self, attrs=f'class="table {BULMA}"'):
+    def display(self, attrs=f'class="table {BULMA}"'):
         return display(
             HTML(j2h.convert(json=remap(self, visit=visit), table_attributes=attrs))
         )
 
 
+class Table(pd.DataFrame):
+    def display(self):
+        return self.plot(**self.attrs)
+
+
 class Attachment(dict):
-    def pretty(self):
-        content = b64decode(self["content"], validate=True)
+    def decode(self):
+        return b64decode(self["content"], validate=True)
+
+    def display(self):
+        content = self.decode()
 
         if self["mime"].startswith("image/"):
             return Image(content)
@@ -169,6 +178,12 @@ class Attachment(dict):
         Path(self["name"]).write_bytes(content)
         return FileLink(self["name"])
 
+    def info(self):
+        fields = ["id", "name", "mime", "md5"]
+        info = Dict((k, v) for k, v in self.items() if k in fields)
+        info["size"] = len(self.decode())
+        return info.display()
+
 
 def load_client(apikey=None, headers=None, host=None):
     warnings.warn(
@@ -348,8 +363,8 @@ class Client(SwaggerClient):
 
         df = pd.DataFrame.from_records(
             table["data"], columns=table["columns"], index=table["index"]
-        )
-        df.attrs = table["attrs"]
+        ).apply(pd.to_numeric, errors="ignore")
+        df.index = pd.to_numeric(df.index, errors="ignore")
         labels = table["attrs"].get("labels", {})
 
         if "index" in labels:
@@ -357,7 +372,9 @@ class Client(SwaggerClient):
         if "variable" in labels:
             df.columns.name = labels["variable"]
 
-        return df
+        ret = Table(df)
+        ret.attrs = table["attrs"]
+        return ret
 
     def get_structure(self, sid):
         """Convenience function to get pymatgen structure."""
@@ -668,6 +685,11 @@ class Client(SwaggerClient):
             "attachments": [<pathlib.Path>, ...]
         }
 
+        This function can also be used to update contributions by including the respective
+        contribution `id`s in the above dictionary and only including fields that need
+        updating. Set list entries to `None` for components that are to be left untouched
+        during an update.
+
         Args:
             contributions (list): list of contribution dicts to submit
             skip_dupe_check (bool): skip check for duplicates of identifiers and components
@@ -689,8 +711,27 @@ class Client(SwaggerClient):
         # get existing contributions
         existing = defaultdict(set)
         existing["unique_identifiers"] = True
-        project_names = list(set(c["project"] for c in contributions))
-        initial_total = self.get_number_contributions(project__in=project_names)
+
+        project_names = set()
+        collect_ids = []
+        for idx, c in enumerate(contributions):
+            if "id" in c:
+                collect_ids.append(c["id"])
+            elif "project" in c:
+                project_names.add(c["project"])
+            else:
+                print(f"Fields `project` or `id` missing for contribution #{idx}!")
+                return
+
+        id2project = {}
+        if collect_ids:
+            resp = self.contributions.get_entries(
+                id__in=collect_ids, _fields=["id", "project"]
+            ).result()
+            id2project = {c["id"]: c["project"] for c in resp["data"]}
+            project_names |= set(id2project.values())
+
+        project_names = list(project_names)
 
         if not skip_dupe_check:
             print("get existing contributions ...")
@@ -712,9 +753,10 @@ class Client(SwaggerClient):
         ]
 
         for contrib in tqdm(contributions, leave=False):
-            project_name = contrib["project"]
+            update = "id" in contrib
+            project_name = id2project[contrib["id"]] if update else contrib["project"]
             if (
-                existing[project_name]["unique_identifiers"]
+                not update and existing[project_name]["unique_identifiers"]
                 and contrib["identifier"] in existing[project_name]["identifiers"]
             ):
                 continue
@@ -728,14 +770,22 @@ class Client(SwaggerClient):
             })
 
             for component in COMPONENTS:
-                contribs[project_name][-1][component] = []
                 elements = contrib.get(component, [])
                 nelems = len(elements)
 
                 if nelems > MAX_ELEMS:
                     raise ValueError(f"Too many {component} ({nelems} > {MAX_ELEMS})!")
 
+                if update and not nelems:
+                    continue  # nothing to update for this component
+
+                contribs[project_name][-1][component] = []
+
                 for idx, element in enumerate(elements):
+                    if update and element is None:
+                        contribs[project_name][-1][component].append(None)
+                        continue
+
                     is_structure = isinstance(element, Structure)
                     is_table = isinstance(element, pd.DataFrame)
                     is_attachment = isinstance(element, Path)
@@ -754,6 +804,7 @@ class Client(SwaggerClient):
                         if not dct.get("charge"):
                             del dct["charge"]
                     elif is_table:
+                        element.fillna('', inplace=True)
                         element.index = element.index.astype(str)
                         for col in element.columns:
                             element[col] = element[col].astype(str)
@@ -811,9 +862,8 @@ class Client(SwaggerClient):
                         msg = f"Duplicate in {project_name}: {contrib['identifier']} {dct['name']}"
                         raise ValueError(msg)
 
-                    if not dupe:
-                        digests[project_name][component].add(digest)
-                        contribs[project_name][-1][component].append(dct)
+                    digests[project_name][component].add(digest)
+                    contribs[project_name][-1][component].append(dct)
 
         # submit contributions
         if contribs:
@@ -833,16 +883,31 @@ class Client(SwaggerClient):
                         data=json.dumps(chunk).encode("utf-8"),
                     )
 
+                def put_future(cdct):
+                    pk = cdct.pop("id")
+                    return session.put(
+                        f"{self.url}/contributions/{pk}/",
+                        headers=headers,
+                        data=json.dumps(cdct).encode("utf-8"),
+                    )
+
                 for project_name in project_names:
                     ncontribs = len(contribs[project_name])
                     total += ncontribs
+                    start = datetime.utcnow()
 
                     while contribs[project_name]:
-                        futures = [
-                            post_future(chunk) for chunk in chunks(
-                                contribs[project_name], n=per_page
-                            )
-                        ]
+                        futures = []
+                        for chunk in chunks(contribs[project_name], n=per_page):
+                            post_chunk = []
+                            for c in chunk:
+                                if "id" in c:
+                                    futures.append(put_future(c))
+                                else:
+                                    post_chunk.append(c)
+
+                            if post_chunk:
+                                futures.append(post_future(post_chunk))
 
                         self._run_futures(futures, total=ncontribs)
 
@@ -857,16 +922,14 @@ class Client(SwaggerClient):
                             if not existing[project_name]["unique_identifiers"] and retry:
                                 print("Please resubmit failed contributions manually.")
 
-                self.load()
-                final_total = self.get_number_contributions(project__in=project_names)
-                submitted_total = final_total - initial_total
+                end = datetime.utcnow()
+                updated_total = self.get_number_contributions(
+                    project__in=project_names, last_modified__gt=start, last_modified__lt=end
+                )
                 toc = time.perf_counter()
                 dt = (toc - tic) / 60
-                print(f"It took {dt:.1f}min to submit {submitted_total} contributions.")
-
-                if submitted_total < total:
-                    failed_total = total - submitted_total
-                    print(f"{failed_total} contributions failed to submit. Check errors/warnings.")
+                self.load()
+                print(f"It took {dt:.1f}min to submit {updated_total} contributions.")
         else:
             print("Nothing to submit.")
 
@@ -882,8 +945,10 @@ class Client(SwaggerClient):
                 if status in {200, 201, 400, 401, 404, 500, 502}:
                     resp = response.json()
                     if status in {200, 201}:
-                        if total:
-                            cnt = len(resp["data"]) if "data" in resp else resp["count"]
+                        if total and "data" in resp:
+                            cnt = len(resp["data"])
+                        elif total and "count" in resp:
+                            cnt = resp["count"]
                         else:
                             cnt = 1
                         pbar.update(cnt)
